/*  Log.h

    Mark Woolrich, FMRIB Image Analysis Group

    Copyright (C) 1999-2000 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

#include <iostream>
#include <fstream>
#include <string>
#include "newmatap.h"
#include "newmatio.h"

using namespace NEWMAT;
namespace UTILS{

#if !defined(__Log_h)
#define __Log_h
  
  class Log
    {
    public:
      static Log& getInstance();
      ~Log() { delete logger; }

      int establishDir(const string& name); 
      void setDir(const string& name); 
      void setLogFile(const string& name) {logfilename = name;}
      const string& getDir() const { return dir; }
      
      void out(const string& p_fname, const Matrix& p_mat, bool p_addMatrixString = true);
      void out(const string& p_fname, const RowVector& p_mat);
      void out(const string& p_fname, const ColumnVector& p_mat);
      void in(const string& p_fname, Matrix& p_mat);
      void in(const string& p_fname, Matrix& p_mat, const int numRows, const int numWaves);
      void in(const string& p_fname, RowVector& p_mat);
      void in(const string& p_fname, ColumnVector& p_mat);

      ofstream& str() { return logfileout; }
      
    private:
      Log() {}
      
      const Log& operator=(Log&);
      Log(Log&);
      
      static Log* logger;
      string dir;
      ofstream logfileout;
      string logfilename;
    };
  
  inline void Log::out(const string& p_fname, const Matrix& p_mat, bool p_addMatrixString)
    {
      ofstream out;
      out.open((dir + "/" + p_fname).c_str(), ios::out);

      if(p_addMatrixString)
	out << "/Matrix" << endl;

      out << p_mat;
      out.close();
    }
  
  inline void Log::out(const string& p_fname, const ColumnVector& p_mat)
    {
      ofstream out;
      out.open((dir + "/" + p_fname).c_str(), ios::out);
      out << p_mat;
      out.close();
    }

  inline void Log::out(const string& p_fname, const RowVector& p_mat)
    {
      ofstream out;
      out.open((dir + "/" + p_fname).c_str(), ios::out);
      out << p_mat;
      out.close();
    }

  inline void Log::in(const string& p_fname, Matrix& p_mat)
    {
      ifstream in;
      in.open((dir + "/" + p_fname).c_str(), ios::in);
      
      if(!in)
	{
	  cerr << "Unable to open " << p_fname << endl;
	  throw;
	}

      int numWaves = 0;
      int numPoints = 0;
       
      string str;

      while(true)
	{
	  in >> str;
	  if(str == "/Matrix")
	    break;
	  else if(str == "/NumWaves")
	    {
	      in >> numWaves;
	    }
	  else if(str == "/NumPoints" || str == "/NumContrasts")
	    {
	      in >> numPoints;
	    }
	}

      if(numWaves != 0)
	{
	  p_mat.ReSize(numPoints, numWaves);
	}
      else
	{
	  numWaves = p_mat.Ncols();
	  numPoints = p_mat.Nrows();
	}
      
      for(int i = 1; i <= numPoints; i++)
	{
	  for(int j = 1; j <= numWaves; j++)    
	    {
	      in >> p_mat(i,j);
	    }
	  }

      in.close();
    }

  inline void Log::in(const string& p_fname, Matrix& p_mat, const int numWaves, const int numPoints)
    {
      ifstream in;
      in.open((dir + "/" + p_fname).c_str(), ios::in);
      
      if(!in)
	{
	  cerr << "Unable to open " << p_fname << endl;
	  throw;
	}

      p_mat.ReSize(numWaves, numPoints);

      for(int i = 1; i <= numPoints; i++)
	{
	  for(int j = 1; j <= numWaves; j++)    
	    {
	      in >> p_mat(i,j);
	    }
	  }

      in.close();
    }

  inline void Log::in(const string& p_fname, ColumnVector& p_mat)
    {
      int size = p_mat.Nrows();

      ifstream in;
      in.open((dir + "/" + p_fname).c_str(), ios::in);
      
      if(!in)
	{
	  string err("Unable to open ");
	  err =  err +  p_fname;
	  throw Exception(err.c_str());
	}

      for(int i = 1; i <= size; i++)
	{
	  in >> p_mat(i);
	}

      in.close();
    }

  inline void Log::in(const string& p_fname, RowVector& p_mat)
    {
      int size = p_mat.Ncols();

      ifstream in;
      in.open((dir + "/" + p_fname).c_str(), ios::in);
      
      if(!in)
	{
	  string err("Unable to open ");
	  err =  err +  p_fname;
	  throw Exception(err.c_str());
	}

      for(int i = 1; i <= size; i++)
	{
	  in >> p_mat(i);
	}

      in.close();
    }

  inline Log& Log::getInstance(){
    if(logger == NULL)
      logger = new Log();
  
    return *logger;
  }

#endif

}


