/*  mcflirt.cc - Motion Correction FLIRT
    
    Peter Bannister, FMRIB Image Analysis Group
    
    Copyright (C) 1999-2001 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */


#include <math.h>
#include <iostream.h>
#include <iomanip>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <strstream>

#include "mjimage.h"
#include "generalio.h"
#include "interpolation.h"
#include "miscimfns.h"
#include "optimise.h"
#include "costfns.h"
#include "newmatap.h"

#include "Globaloptions.h"
#include "Log.h"

using namespace MISCMATHS;
using namespace MISCIMFNS;
using namespace INTERPOLATION;
using namespace COSTFNS;
using namespace NEWMAT;
using namespace MJIMAGE;
using namespace GENERALIO;
using namespace UTILS;



  int affine_transform_sinc(volume& vout, const volume& vin, const Matrix& aff)
    {
      // does the transform with an effective one voxel padding of
      // vin around the edges (achieves this without changing or
      // copying vin)

      // takes the volume (vin) and applies a spatial transform, given
      //  by the 4x4 matrix (aff) in WORLD COORDINATES
      // the result is a new volume (vout)


      // Do everything in practice via the inverse transformation
      // That is, for every point in vout, calculate the pre-image in
      //  vin to which it corresponds, and interpolate vin to get the
      //  value there.
      // Also, the sampling transformations must be accounted for:
      //     T_vox1->vox2 = (T_samp2)^-1 * T_world * T_samp1
      Matrix iaffbig = vin.sampling_matrix().i() * aff.i() *
	                     vout.sampling_matrix();  
      Matrix iaff=iaffbig.SubMatrix(1,3,1,3);
      unsigned int xb2=vout.columns()-1, yb2=vout.rows()-1, zb2=vout.slices()-1;

      float a11=iaff(1,1), a12=iaff(1,2), a13=iaff(1,3), a14=iaffbig(1,4),
	a21=iaff(2,1), a22=iaff(2,2), a23=iaff(2,3), a24=iaffbig(2,4),
	a31=iaff(3,1), a32=iaff(3,2), a33=iaff(3,3), a34=iaffbig(3,4), o1,o2,o3;
  
  // cout << "Bounds are " << xb << " " << yb << " " << zb << endl;
  // The matrix algebra below has been hand-optimized from
  //  [o1 o2 o3] = a * [x y z]  at each iteration
      for (unsigned int z=0; z<=zb2; z++) { 
	for (unsigned int x=0; x<=xb2; x++) { 
	  o1=x*a11 + z*a13 + a14;  // y=0
	  o2=x*a21 + z*a23 + a24;  // y=0
	  o3=x*a31 + z*a33 + a34;  // y=0
	  for (unsigned int y=0; y<=yb2; y++) {
	    vout(x,y,z) = sinc_interpolation(vin,o1,o2,o3);

	    o1 += a12;
	    o2 += a22;
	    o3 += a32;
	  }
	}
      }
      return 0;
    }



//------------------------------------------------------------------------//

// OPTIMISATION SUPPORT


int vector2affine(const ColumnVector& params, int n, const ColumnVector& centre,
 		  Matrix& aff)
{
  if (n<=0) return 0;
  // order of parameters is 3 rotation + 3 translation + 3 scales + 3 skews
  // angles are in radians
  
  switch (Globaloptions::getInstance().anglerep) 
    {
    case Euler:
      compose_aff(params,n,centre,aff,construct_rotmat_euler);
      break;
    case Quaternion:
      compose_aff(params,n,centre,aff,construct_rotmat_quat);
      break;
    default:
      cerr << "Invalid Rotation Representation" << endl;
      return -1;
    }
  return 0;
}  


int vector2affine(const ColumnVector& params, int n, Matrix& aff)
{
  return vector2affine(params,n,Globaloptions::getInstance().impair->testvol.cog(),aff);
}


 int vector2affine(const float params[], int n, Matrix& aff)
 {
   ColumnVector vparams(12);
   for (int i=1; i<=n; i++) {
     vparams(i) = params[i];
   }
   return vector2affine(vparams,n,aff);
 }


 int affmat2vector(const Matrix& aff, int n, const ColumnVector& centre,
 		  ColumnVector& params)
 {
   switch (Globaloptions::getInstance().anglerep) 
     {
     case Euler:
       decompose_aff(params,aff,centre,rotmat2euler);
       break;
     case Quaternion:
       decompose_aff(params,aff,centre,rotmat2quat);
       break;
     default:
       cerr << "Invalid Rotation Representation" << endl;
     }
   return 0;
 }


 int affmat2vector(const Matrix& aff, int n, ColumnVector& params)
 {
   return affmat2vector(aff,n,Globaloptions::getInstance().impair->testvol.cog(),params);
 }


 void set_param_basis(Matrix &parambasis, int no_params)
 {
   parambasis = 0.0;
   for (int i=1; i<=no_params; i++) {
     parambasis(i,i)=1.0;
   }
 }

 void set_param_tols(ColumnVector &param_tol, int no_params)
 {
   Globaloptions& globalopts = Globaloptions::getInstance();
       // Tolerances are: 0.57 degrees (0.005 radians), 0.2mm translation
       //    0.02 scale and 0.001 skew
   float diagonal[12]={0.005, 0.005, 0.005, 0.2, 0.2, 0.2, 0.002, 0.002, 0.002,
   		      0.001, 0.001, 0.001};
   if (param_tol.Nrows()<no_params) {
     param_tol.ReSize(no_params);
   }
   for (int i=1; i<=no_params; i++) {
     param_tol(i)=diagonal[i-1];
   }
   for (int i=1; i<4; i++)
     param_tol(i) /= globalopts. rot_param;
 }


 void initialise_params(ColumnVector& params)
 {
   Real paramsf[12] = {0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0};
   params << paramsf;
 }



 void powell_opt(ColumnVector& params, int no_params, ColumnVector& param_tol, 
 		int &no_its, float &fans, float (*costfunc)(const ColumnVector&), 
 		int itmax)
 {
   // sets up the initial parameters and calls the powell optimisation routine
   if (params.MaximumAbsoluteValue() < 0.001)  initialise_params(params);
   { 
     Matrix affmattst(4,4);
     vector2affine(params,no_params,affmattst);
   }
   Matrix parambasis(no_params,no_params);
   set_param_basis(parambasis,no_params);
   float ptol[13];
   for (int i=1; i<=no_params; i++) { ptol[i] = param_tol(i); }
  
   // the optimisation call
   //powell(params,parambasis,no_params,ptol,no_its, fans, costfunc, itmax);
   fans = MISCMATHS::optimise(params,no_params,param_tol,costfunc,no_its,itmax,
 			      Globaloptions::getInstance().boundguess);
 }


 void optimise(ColumnVector& params, int no_params, ColumnVector& param_tol, 
 	      int &no_its, float &fans, float (*costfunc)(const ColumnVector &), 
 	      int itmax=4)
 {
   powell_opt(params,no_params,param_tol,no_its,fans,costfunc,itmax);
 }

 ////////////////////////////////////////////////////////////////////////////

 // OPTIMISATION SUPPORT (cost function interfaces)


 float costfn(const Matrix& uninitaffmat)
 {
   Tracer tr("costfn");
   Matrix affmat = uninitaffmat * Globaloptions::getInstance().initmat;  // apply initial matrix
   float retval = 0.0;
   Globaloptions& globalopts = Globaloptions::getInstance();

   switch (globalopts.maincostfn) 
     {
     case NormCorr:
       if (globalopts.smoothsize > 0.0) {
	 retval = 1.0 - fabs(normcorr_smoothed(globalopts.impair,affmat)); // MAXimise corr
       } else {
	 retval = 1.0 - fabs(normcorr(globalopts.impair,affmat)); // MAXimise corr
       }
       break;
     case NormCorrSinc:
       retval = 1.0 - fabs(normcorr_smoothed_sinc(globalopts.impair,affmat)); // MAXimise corr
       break;
     case CorrRatio:
       if (globalopts.smoothsize > 0.0) {
 	retval = 1.0 - corr_ratio_smoothed(globalopts.impair,affmat);  // MAXimise corr
       } else {
 	retval = 1.0 - corr_ratio(globalopts.impair,affmat);  // MAXimise corr
       }
       break;
     case Woods:
       retval = woods_fn(globalopts.impair,affmat);  // minimise variance/mean
       break;
     case MutualInfo:
       retval = -mutual_info(globalopts.impair,affmat);  // MAXimise info
       break;
     case NormMI:
       retval = -normalised_mutual_info(globalopts.impair,affmat);  // MAXimise
       break;
     case LeastSq:
       retval = leastsquares(globalopts.impair, affmat); // MINimise squared difference
       break;
     default:
       cerr << "Invalid cost function type" << endl;
       break;
     }
   return retval;
 }


 float costfn(const ColumnVector& params)
 {
   Tracer tr("costfn");
   Matrix affmat(4,4);
   vector2affine(params,Globaloptions::getInstance().no_params,affmat);
   float retval = costfn(affmat);
   return retval;
 }
  
 /*
 float costfn(float params[])
 {
   Tracer tr("costfn");
   Matrix affmat(4,4);
   vector2affine(params,Globaloptions::getInstance().no_params,affmat);
   float retval = costfn(affmat);
   if (Globaloptions::getInstance().verbose>=5) {
     cout << Globaloptions::getInstance().impair->count++ << " : ";
     cout << retval << " :: ";
     for (int i=1; i<=Globaloptions::getInstance().no_params; i++) cout << params[i] << " ";
     cout << endl;
   }
   return retval;
 }
 */

 //------------------------------------------------------------------------//

 float estimate_scaling(const volume& vol) {
   Tracer tr("estimate_scaling");
   return Min(Min(vol.getx(),vol.gety()),vol.getz());
 }

 float estimate_scaling() {
   Tracer tr("estimate_scaling");
   return estimate_scaling(Globaloptions::getInstance().impair->refvol);
 }

 ////////////////////////////////////////////////////////////////////////////

 int optimise_strategy1(Matrix& matresult, float& fans, int input_dof, 
 		       int max_iterations, float new_tolerance)
 {
   Tracer tr("optimise_strategy1");
   // the most basic strategy - just do a single optimisation run at the
   //  specified dof
   int dof=input_dof;
   if (dof<6) { 
     cerr << "Erroneous dof " << dof << " : using 6 instead\n"; 
     dof=6; 
   }
   if (dof>12) {
     cerr << "Erroneous dof " << dof << " : using 12 instead\n"; 
     dof=12;
   }

   ColumnVector params(12), param_tol(12);
   int no_its=0;
   Globaloptions::getInstance().no_params = dof;
   set_param_tols(param_tol,12);
   param_tol = param_tol * new_tolerance;
   
   affmat2vector(matresult,dof,params);
   optimise(params,dof,param_tol,no_its,fans,costfn,max_iterations);
   
   //cerr << no_its << endl;
   vector2affine(params,dof,matresult);
   return no_its;
 }  

 ////////////////////////////////////////////////////////////////////////////


 void usroptimise(Matrix &matresult, int usrdof, int usrmaxitn, float new_tolerance)
 {
   Tracer tr("usroptimise");
   // OPTIMISE
   int dof = Min(Globaloptions::getInstance().dof,usrdof);
  
   float costval=0.0;
   optimise_strategy1(matresult,costval,dof,usrmaxitn, new_tolerance);
 

   // matresult is the desired solution (cost is costval)
 }


void usrsetscale(volume& newrefvol, volume& newtestvol, int usrscale) {
   Tracer tr("usrsetscale");
   imagepair *globalpair=0;
   if (Globaloptions::getInstance().impair != 0) 
     delete Globaloptions::getInstance().impair;
   globalpair = new imagepair(newrefvol,newtestvol);
   globalpair->set_no_bins(Globaloptions::getInstance().no_bins/usrscale);
   Globaloptions::getInstance().impair = globalpair;
 }


 void g_smooth(volume& testvol) {
   Tracer tr("g_smooth");
   volume result;
   
   result = testvol;
   volume g_kern = gaussian_kernel(1.933, 8);
   mirror_convolve(testvol, result, g_kern);
   testvol = result;
 }

 ////////////////////////////////////////////////////////////////////////////

 void correct(const int direction, volume& reference_volume, tvolume& timeseries, const float scaling, const float new_tolerance, vector<Matrix>& mat_array_in, vector<Matrix>& mat_array_out, int mean_cond) {
   Tracer tr("correct");
   Globaloptions& globalopts = Globaloptions::getInstance();
   volume testvol, newtestvol, refvol;
   Matrix offsettrans(4,4), finalmat(4,4);

   Identity(offsettrans);
   int i=globalopts. refnum + direction;
   int stop = -1 - mean_cond;

   refvol = reference_volume;

   while ( i != (direction == 1 ? globalopts. no_volumes : stop)) {
     if (!globalopts. no_reporting) 
       cerr << "[" << i << "]";
     testvol = timeseries[i];

     if (globalopts. edgeflag){
       if (!globalopts. no_reporting) 
	 cerr <<"Calculating contour image for volume [" << i << "]" << endl;
       fixed_edge_detect(testvol, 15000);  
     } else if (globalopts. gdtflag){
       if (!globalopts. no_reporting) 
	 cerr <<"Calculating gradient image for volume [" << i << "]" << endl;
       volume gtempvol = testvol;
       gradient(gtempvol, testvol);
     }
    
     offsettrans = mat_array_in[i];
     usrsetscale(reference_volume,testvol,(int) scaling);
     usroptimise(offsettrans,globalopts.dof,1,new_tolerance);
     {
       finalmat = offsettrans * globalopts.initmat;
       if (globalopts. tmpmatflag) {
	 mat_array_out[i] = finalmat;
       }
     }

     i += direction;
     /*
     if ((scaling == 8.0) && (i < globalopts. no_volumes - 1) && (i > -1))
       // if first scaling level, use previous result to initialise next transformation search
       mat_array_in[i] = finalmat;
     */
   }
 }

 void concatenate_mats(int* mat_index, vector<Matrix>& mat_array, const volume& refvol){
   Tracer tr("concatenate_mats");
   Globaloptions& globalopts = Globaloptions::getInstance();
   Log& logger = Log::getInstance();
   Matrix level0(4,4), level1(4,4), level2(4,4), product(4,4);
   ColumnVector param_vec(12);
   param_vec = 0;
   
   string filename = globalopts. outputfname + ".par";
   ofstream outfile;

   if (globalopts. plotflag)
     outfile. open(filename.c_str());
   
   string pathname = globalopts. inputfname;
   find_pathname(pathname);
   if (globalopts. matflag) {
     if (logger.establishDir(globalopts. outputfname + ".mat") != 0){
       cerr << "Error! Could not create directory: " << pathname << logger. getDir() << ". No write permission" << endl;
       globalopts. tmpmatflag = 0;
     }
   }

   for (int i = 0; i < globalopts. no_volumes; i++){
     if (i == globalopts. refnum){
       char strc[50];
       ostrstream osc(strc,50);
       product = Identity(4);
       osc << "MAT_" << setw(4) << setfill('0') << i  << '\0';
       if (globalopts. tmpmatflag)
	 logger.out(strc, product, false);
       if (globalopts. plotflag) {
	 //decompose_aff(param_vec, product, refvol.cog, rotmat2euler);
	 decompose_aff(param_vec, product, refvol.cog(), rotmat2euler);
	 if (!outfile) {
	   cerr << "error: unable to open output file!\n";
	   exit(0);
	 }
	 outfile << param_vec(1) << "  " << param_vec(2) << "  " 
		 << param_vec(3) << "  " << param_vec(4) << "  " 
		 << param_vec(5) << "  " << param_vec(6) << "  " << endl;
       }
       
     } else {
       char strcP[50];
       ostrstream oscP(strcP,50);
       oscP << "MAT_" << setw(4) << setfill('0') << i  << '\0';
       if (globalopts. matflag)
	 logger.out(strcP, mat_array[i], false);
       if (globalopts. plotflag){
	 decompose_aff(param_vec, mat_array[i], refvol.cog(), rotmat2euler);
	 if (!outfile) {
	   cerr << "error: unable to open output file!\n";
	   exit(0);
	 }
	 outfile << param_vec(1) << "  " << param_vec(2) << "  " 
		 << param_vec(3) << "  " << param_vec(4) << "  " 
		 << param_vec(5) << "  " << param_vec(6) << "  " << endl;
       }
     }
   }
 }

 void eval_costs(volume& refvol, tvolume& timeseries, vector<Matrix>& mat_array, float current_scale) {
   Globaloptions& globalopts = Globaloptions::getInstance();
   ofstream outfile;
   
   outfile. open("/usr/people/prb/medx/motion/releasetest/costs.txt");
   for (int i=0; i < globalopts. no_volumes; i++){
     usrsetscale(refvol, timeseries[i], (int)current_scale);
     outfile << costfn(mat_array[i]) << endl;
   }
 }

 
 void run_and_save_stats(const tvolume& timeseries) {
   Tracer tr("run_and_save_stats");
   Globaloptions& globalopts = Globaloptions::getInstance();
   volume variancevol, meanvol, sigmavol;
  
   int vmax = timeseries.getSize();

   meanvol = timeseries[0];
   variancevol = timeseries[0];
   sigmavol = timeseries[0];

   meanvol.reinitialize(timeseries[0].xsize(), timeseries[0].ysize(), (timeseries[0].zsize() - 2));
   variancevol.reinitialize(meanvol);
   sigmavol.reinitialize(meanvol);

   meanvol = 0.0;
   variancevol = 0.0;
   sigmavol = 0.0;

   // calculate the mean value and variance at each voxel
    for (int x=0; x< timeseries[0].xsize(); x++) {
     for (int y=0; y< timeseries[0].ysize(); y++) {
       for (int z=0; z< (timeseries[0].zsize()-2); z++) {
	 for (int i=0; i< vmax; i++) 
 	  meanvol(x,y,z) = meanvol(x,y,z) + timeseries[i](x,y,z+1);
 	meanvol(x,y,z) = meanvol(x,y,z)/(float)vmax;
 	for (int i=0; i< vmax; i++)
	  variancevol(x,y,z) = variancevol(x,y,z) + (timeseries[i](x,y,z+1) - meanvol(x,y,z))*(timeseries[i](x,y,z+1) - meanvol(x,y,z));
	variancevol(x,y,z) = variancevol(x,y,z)/((float)(vmax - 1));
 	sigmavol(x,y,z) = sqrt(variancevol(x,y,z));
       }
     }
   }

   if (!globalopts. no_reporting) 
     cerr <<"Saving mean volume... " << endl;
   save_volume(meanvol, "mean_"+globalopts.outputfname);
  
   if (!globalopts. no_reporting) 
     cerr <<"Saving variance volume... " << endl;
   save_volume(variancevol, "variance_"+globalopts.outputfname);

   if (!globalopts. no_reporting) 
     cerr <<"Saving standard deviation volume... " << endl;
   save_volume(sigmavol, "sigma_"+globalopts.inputfname);
 }



int main (int argc,char** argv)
{
  Tracer tr("main");
  tvolume timeseries;
  volume refvol, anisorefvol, testvol, meanvol;
  Globaloptions& globalopts = Globaloptions::getInstance();
  short dtype;
  float current_scale=8.0, new_tolerance=0.8;
  int mat_index[3];
  ColumnVector centre(3);
  vector<Matrix> mat_array0, mat_array1, mat_array2;
  int mean_cond = 0;

  globalopts.parse_command_line(argc, argv);
  if (!globalopts. no_reporting) cerr << endl << "McFLIRT v 1.1.18 - fMRI affine motion correction tool" << endl << 
    "This software is currently beta." << endl << "Please send any comments/ queries to prb@fmrib.ox.ac.uk" << endl << endl;

  if (!globalopts. no_reporting) cerr <<"Reading time series... " << endl;
  read_series(timeseries, globalopts.inputfname, dtype);
  globalopts.datatype = dtype;
  globalopts. no_volumes = timeseries.getSize();
  
  for (int vol_count = 0; vol_count < globalopts. no_volumes; vol_count++) {
    mat_array0.push_back(Identity(4));
    mat_array1.push_back(Identity(4));
    mat_array2.push_back(Identity(4));
  }

  if (globalopts. refnum == -1) globalopts. refnum = globalopts. no_volumes/2;

  for (int mean_its = 0; mean_its < 1 + globalopts. meanvol; mean_its++) {
    if (globalopts. no_stages>=1) {
      if (!globalopts. no_reporting) cerr <<"first iteration - 8mm scaling, set tolerance" << endl;
      new_tolerance=8*0.2*0.5; current_scale=8.0; mat_index[0] = (int) (new_tolerance*current_scale);
      
      if (mean_its == 0) {
	anisorefvol = timeseries[globalopts. refnum];
      } else { //2nd pass - generate mean volume, clear mat_array0
	meanvol = timeseries[0];	
	for (int x=0; x< meanvol. xsize(); x++) {
	  for (int y=0; y< meanvol. ysize(); y++) {
	    for (int z=0; z< meanvol .zsize(); z++) {
	      for (int i = 1; i < globalopts. no_volumes; i++)	  
		meanvol(x,y,z) = meanvol(x,y,z) + timeseries[i](x,y,z);
	      meanvol(x,y,z) = meanvol(x,y,z)/(float)globalopts. no_volumes;    
	    }
	  }
	}
	save_volume(meanvol, globalopts.outputfname + "_mean_reg");
	
	anisorefvol = meanvol;
	globalopts. refnum = -1; // to ensure stopping condition in ::correct subroutine
	for (int i=0; i < globalopts. no_volumes; i++)
	  mat_array0[i] = Identity(4);
	mean_cond = 1;
      }
      
      if (!globalopts. no_reporting)  cerr <<"Rescaling reference volume [" << globalopts. refnum << "] to " 
					   << current_scale << " mm pixels" << endl;
      isotropic_rescale(refvol,anisorefvol,current_scale);
      
      if (globalopts. edgeflag){
	if (!globalopts. no_reporting) cerr <<"Calculating contour image for reference volume" << endl;
	fixed_edge_detect(refvol, 15000);  
	if (!globalopts. no_reporting) cerr <<"Saving contour reference volume... " << endl;
	save_volume(refvol, "crefvol_"+globalopts.outputfname);
      } else if (globalopts. gdtflag){
	if (!globalopts. no_reporting) cerr <<"Calculating gradient image for reference volume" << endl;
	volume gtempvol = refvol; gradient(gtempvol, refvol);
	if (!globalopts. no_reporting) cerr <<"Saving gradient reference volume... " << endl;
	save_volume(refvol, "grefvol_"+globalopts.outputfname);
      }
      
      Identity(globalopts.initmat);
      if (!globalopts. no_reporting) cerr <<"Registering volumes ... ";
      correct(1, refvol, timeseries, current_scale, new_tolerance, mat_array0, mat_array1, mean_cond);
      correct(-1, refvol, timeseries, current_scale, new_tolerance, mat_array0, mat_array1, mean_cond);
    } else {
      for (int i=0; i<globalopts.no_volumes; i++)  mat_array1[i] = mat_array0[i];
    }
    
    if (globalopts. no_stages>=2) {
      if (!globalopts. no_reporting) cerr <<endl << "second iteration - drop to 4mm scaling" << endl;
      new_tolerance=4*0.2; current_scale=4.0; mat_index[1] = (int) (new_tolerance*current_scale);
      
      if (!globalopts. no_reporting) cerr <<"Rescaling reference volume [" << globalopts. refnum << "] to " 
					  << current_scale << " mm pixels" << endl;
      isotropic_rescale(refvol,anisorefvol,current_scale);
      
      if (globalopts. edgeflag){
	if (!globalopts. no_reporting) cerr <<"Calculating contour image for reference volume" << endl;
	fixed_edge_detect(refvol, 15000);  
	if (!globalopts. no_reporting) cerr <<"Saving contour reference volume... " << endl;
	save_volume(refvol, "crefvol_"+globalopts.outputfname);
      } else if (globalopts. gdtflag){
	if (!globalopts. no_reporting) cerr <<"Calculating gradient image for reference volume" << endl;
	volume gtempvol = refvol;
	gradient(gtempvol, refvol);
	if (!globalopts. no_reporting) cerr <<"Saving gradient reference volume... " << endl;
	save_volume(refvol, "grefvol_"+globalopts.outputfname);
      }
      
      if (!globalopts. no_reporting) cerr <<"Registering volumes ... ";
      correct(1, refvol, timeseries, current_scale, new_tolerance, mat_array1, mat_array2, mean_cond);
      correct(-1, refvol, timeseries, current_scale, new_tolerance, mat_array1, mat_array2, mean_cond);
    } else {
      for (int i=0; i<globalopts.no_volumes; i++)  mat_array2[i] = mat_array1[i];
    }
    
    
    if (globalopts. no_stages>=3) {
      if (!globalopts. no_reporting) cerr << endl << "third (final) iteration - still at 4mm scaling, eighth tolerance" << endl;
      new_tolerance = 0.1; mat_index[2] = (int) (new_tolerance*current_scale);
      
      if (!globalopts. no_reporting) cerr <<"Registering volumes ... ";
      correct(1, refvol, timeseries, current_scale, new_tolerance, mat_array2, mat_array1, mean_cond);
      correct(-1, refvol, timeseries, current_scale, new_tolerance, mat_array2, mat_array1, mean_cond);
    } else {
      for (int i=0; i<globalopts.no_volumes; i++)  mat_array1[i] = mat_array2[i];
    }
  }
      
  mean_cond = 0;

  if (globalopts. no_stages>=4) {
    if (!globalopts. no_reporting) cerr << endl << "fourth (really final) iteration - still at 4mm scaling, eighth tolerance, but with new improved interpolation" << endl;
    
    if (!globalopts. no_reporting) cerr <<"Registering volumes ... ";
    if (globalopts.maincostfn == NormCorr)  globalopts.maincostfn = NormCorrSinc;
    correct(1, refvol, timeseries, current_scale, new_tolerance, mat_array1, mat_array0, mean_cond);
    correct(-1, refvol, timeseries, current_scale, new_tolerance, mat_array1, mat_array0, mean_cond);
  } else {
    for (int i=0; i<globalopts.no_volumes; i++)  mat_array0[i] = mat_array1[i];
  }

  Matrix init_trans(4,4);
  if (globalopts.init_transform.size()<1) {
    init_trans = Identity(4);
  } else {
    read_matrix(init_trans,globalopts.init_transform,
		timeseries[globalopts.refnum]);
  }

  // interpolate with final transform values
  for (int i=0; i < globalopts. no_volumes; i++){
    testvol = timeseries[i];
    filled_affine_transform(testvol,timeseries[i],
			    mat_array0[i]*init_trans);
    timeseries[i] = testvol;
  }

  if (!globalopts. no_reporting) cerr << endl << "Saving motion corrected time series... " << endl;
  save_series(timeseries, globalopts.outputfname, globalopts.datatype);

  if (globalopts. statflag) run_and_save_stats(timeseries);
  if (globalopts. tmpmatflag) concatenate_mats(mat_index, mat_array0, timeseries[globalopts. refnum]);
  if (globalopts. costmeas) eval_costs(refvol, timeseries, mat_array0, current_scale);

}









