#{{{ copyright and setup

#   MELODIC - Multivariate Exploratory Linear Optimized Decomposition into 
#             Independent Components
#
#   Christian Beckmann and Stephen Smith, FMRIB Image Analysis Group
#
#   Copyright (C) 2001 University of Oxford
#
#   Part of FSL - FMRIB's Software Library
#   WWW:      http://www.fmrib.ox.ac.uk/fsl
#   Email:    fsl@fmrib.ox.ac.uk
#   
#   Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
#   Imaging of the Brain), Department of Clinical Neurology, Oxford
#   University, Oxford, UK
#   
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or (at
#   your option) any later version.
#   
#   This program is distributed in the hope that it will be useful, but in
#   order that the University as a charitable foundation protects its
#   assets for the benefit of its educational and research purposes, the
#   University makes clear that no condition is made or to be implied, nor
#   is any warranty given or to be implied, as to the accuracy of FSL, or
#   that it will be suitable for any particular purpose or for use under
#   any specific conditions.  Furthermore, the University disclaims all
#   responsibility for the use which is made of FSL.  See the GNU General
#   Public License for more details.
#   
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#   USA

source [ file dirname [ info script ] ]/fslstart.tcl

set VARS(history) {}

set VERSION [ exec ${FSLDIR}/bin/melodic -V | awk {{ print $3 }} - ]

#}}}
#{{{ main GUI

proc melodic { w } {

    #{{{ vars and setup

global vars FSLDIR INMEDX argc argv PWD VERSION entries

toplevel $w
wm title $w "MELODIC - v$VERSION"
wm iconname $w "MELODIC"
wm iconbitmap $w @${FSLDIR}/tcl/fmrib.xbm

frame $w.f

set vars(npts) 0

#}}}
    #{{{ dim

frame $w.f.n

label $w.f.n.npts -text "Total volumes $vars(npts)"

set vars(dim) $vars(npts)

tixControl $w.f.n.dim -label "    Output components  " -variable vars(dim) \
	-step 1 -min 1 -max $vars(npts) -selectmode immediate -options { entry.width 4 }

pack $w.f.n.npts $w.f.n.dim -in $w.f.n -side left

#}}}
    #{{{ input image (if not in medx)

if { $INMEDX } {

    if { [ what_type 0 ] == 4 } {
	MxGetCurrentPage currentpage
	MxGetPageProperties $currentpage Props
	set entries [ keylget Props Name ]
    }

    frame $w.f.input
    label $w.f.input.label -text "4D input data "
    entry $w.f.input.entry -textvariable entries -width 35
    button $w.f.input.button -text "Select" -command "SelectSinglePage:Dialog $w entries \"melodic:select $w 0\""
    pack $w.f.input.label $w.f.input.entry $w.f.input.button -in $w.f.input -padx 3 -pady 0 -side left

    melodic:select $w 0

} else {
    if { $argc > 0 && [ string length [ lindex $argv 0 ] ] > 0 } {
        set inputname [ file rootname [ lindex $argv 0 ] ].hdr
        if { [ string first / $inputname ] == 0 || [ string first ~ $inputname ] == 0 } {
            set vars(input) $inputname
        } else {
            set vars(input) ${PWD}/$inputname
        }
	melodic:select $w 0
    }

    FSLFileEntry $w.f.input \
            -variable vars(input) \
            -pattern "*.hdr" \
            -directory $PWD \
            -label "4D input data " \
            -title "Select the 4D input data" \
            -width 35 \
            -filterhist VARS(history) \
	    -command "melodic:select $w"
}

pack $w.f.input -in $w.f -side top -padx 5 -pady 5 -anchor w

#}}}
    #{{{ Advanced Options

collapsible frame $w.f.opts -title "Advanced Options"

#{{{ masking

frame $w.f.opts.bet

label $w.f.opts.bet.label -text "Use BET to mask data"

set vars(bet) 1
checkbutton $w.f.opts.bet.yn -variable vars(bet) -command "melodic:updatebet $w"

pack $w.f.opts.bet.label $w.f.opts.bet.yn -in $w.f.opts.bet -side left -padx 5

set vars(thresh) 10

tixControl $w.f.opts.bet.thresh -label "  Threshold % " -variable vars(thresh) \
	-step 1 -min 0 -max 100 -selectmode immediate -options { entry.width 3 }

#}}}
#{{{ variance normalisation

frame $w.f.opts.varnorm

label $w.f.opts.varnorm.label -text "Variance-normalise timecourses"

set vars(varnorm) 1
checkbutton $w.f.opts.varnorm.yn -variable vars(varnorm)

pack $w.f.opts.varnorm.label $w.f.opts.varnorm.yn -in $w.f.opts.varnorm -side left -padx 5

#}}}
#{{{ web report

frame $w.f.opts.report

label $w.f.opts.report.label -text "Create web page report"

set vars(report) 1
checkbutton $w.f.opts.report.yn -variable vars(report)

pack $w.f.opts.report.label $w.f.opts.report.yn -in $w.f.opts.report -side left -padx 5

#}}}

pack $w.f.opts.bet $w.f.opts.varnorm $w.f.opts.report -in $w.f.opts.b -side top -anchor w -pady 5

#}}}

    pack $w.f.n $w.f.opts -in $w.f -side top -padx 5 -pady 5 -anchor w

    #{{{ Button Frame

    frame $w.btns
    frame $w.btns.b -relief raised -borderwidth 1
 
    button $w.apply     -command "melodic:apply $w keep" \
        -text "Go" -width 5
    bind $w.apply <Return> {
        [winfo toplevel %W].apply invoke}
 
    button $w.cancel    -command "melodic:destroy $w" \
        -text "Exit" -width 5
    bind $w.cancel <Return> {
        [winfo toplevel %W].cancel invoke}
 
    button $w.help -command "FmribWebHelp file: ${FSLDIR}/doc/melodic/index.html" \
            -text "Help" -width 5
    bind $w.help <Return> {
        [winfo toplevel %W].help invoke}

    pack $w.btns.b -side bottom -fill x
    pack $w.apply $w.cancel $w.help -in $w.btns.b \
        -side left -expand yes -padx 3 -pady 10 -fill y
 
    pack $w.f $w.btns -expand yes -fill both

#}}}
}

#}}}
#{{{ melodic:updatebet

proc melodic:updatebet { w } {

    global vars

    if { $vars(bet) == 1 } {
	pack forget $w.f.opts.bet.thresh
    } else {
	pack $w.f.opts.bet.thresh -in $w.f.opts.bet -side left -padx 5 -after $w.f.opts.bet.yn
    }
}

#}}}
#{{{ melodic:select

proc melodic:select { w dummy } {

    global vars FSLDIR INMEDX entries

    if { $INMEDX } {

	MxGetCurrentFolder Folder
	MxGetPageByName $Folder $entries pentries
	MxGetPagesFromGroup $pentries GPages
	if { [ info exists GPages ] } {
	    set vars(npts) [ llength $GPages ]
	} else {
	    puts "Warning - you must select a group page"
	}

    } else {
	set vars(npts) [ exec sh -c "${FSLDIR}/bin/avwval $vars(input) dim4" ]
    }

    $w.f.n.npts configure -text "Total volumes $vars(npts)"

    $w.f.n.dim configure -max $vars(npts)
    set vars(dim) [ expr $vars(npts) / 4 ]

    if { $vars(npts) < 2 } {
	puts "Warning - not enough volumes in input data"
    }
}

#}}}
#{{{ melodic:apply

proc melodic:apply { w dialog } {
    global vars HOME FSLDIR INMEDX entries
    
    #{{{ save file if in MEDx

if { $INMEDX } {

    MxGetCurrentFolder Folder
    MxGetPageByName $Folder $entries pentries

    set GroupFullNameTmp $entries

    regsub -all " " $GroupFullNameTmp "" GroupFullNameTmp2
    regsub -all "\"" $GroupFullNameTmp2 "" GroupFullName

    # find directory and filetailroot
    set GroupDirName [ file dirname  $GroupFullName ]
    regsub -all "\\." [ file rootname [ file tail $GroupFullName ] ] "_" GroupFileRoot

    set NotTmp [ string compare /tmp [ string range $GroupDirName 0 3 ] ]
    if { ! [ file writable $GroupDirName ] || $NotTmp == 0 } {
	set GroupName ${HOME}/${GroupFileRoot}
    } else {
	set GroupName ${GroupDirName}/${GroupFileRoot}
    }

    set vars(input) [ new_filename ${GroupName}.hdr ]

    if { [ FSLSaveAs $pentries AVW $vars(input) true ] } {
	MxPause "Please select a valid group page"
	return 1
    }
    
}

#}}}
    #{{{ setup filenames

set INPUT [ file rootname $vars(input) ]

set OUTPUT [ new_filename ${INPUT}.ica ]

if { ! [ file writable [ file dirname $OUTPUT ] ] } {
    set OUTPUT [ new_filename ${HOME}/[ file tail $OUTPUT ] ]
}

#}}}

    set thecommand "${FSLDIR}/bin/melodic -i $INPUT -o $OUTPUT -d $vars(dim)"

    #{{{ masking

if { ! $vars(bet) } {
    set thecommand "$thecommand --nobet -t $vars(thresh)"
}

#}}}
    #{{{ varnorm

if { $vars(varnorm) == 0 } {
    set thecommand "$thecommand --vn"
}

#}}}

    puts $thecommand
    catch { exec sh -c $thecommand } errmsg
    if { $errmsg != "" } {
	puts $errmsg
    } 

    #{{{ create web report

    if { $vars(report) } {
	puts "${FSLDIR}/bin/melodicreport $OUTPUT"
	catch { exec sh -c "${FSLDIR}/bin/melodicreport $OUTPUT" } errmsg
	puts $errmsg
    }

#}}}
    #{{{ if in MEDx load ICA output

if { $INMEDX } {

    MxGetCurrentFolder Folder
    MxOpenImage $Folder ${OUTPUT}/melodic_IC.hdr Image

}

#}}}

    puts Done

    update idletasks

    if {$dialog == "destroy"} {
        melodic:destroy $w
    }
}

#}}}
#{{{ melodic:destroy

# Summary:      Destroys melodic dialog box
proc melodic:destroy { w } {
    destroy $w
}

#}}}
#{{{ call GUI and wait

wm withdraw .
melodic .rename
tkwait window .rename

#}}}
