
/*  MELODIC - Multivariate exploratory linear optimized decomposition into 
              independent components
    
    melodicoptions.cc - class for command line options

    Christian F. Beckmann, FMRIB Image Analysis Group
     
    Copyright (C) 1999-2001 University of Oxford */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <stdio.h>
#include "Log.h"
#include "melodicoptions.h"
#include "newimageall.h"

using namespace Utilities;
using namespace NEWIMAGE;

namespace Melodic {

MelodicOptions* MelodicOptions::gopt = NULL;

  void MelodicOptions::parse_command_line(int argc, char** argv, Log& logger, const string &p_version)
{
  //set version number
  version = p_version;

  // parse once to establish log directory name
  for(int a = options.parse_command_line(argc, argv); a < argc; a++);

  // act on simple command line arguments
  if(help.value())
    {
      print_usage(argc, argv);
      exit(0);
    } 
  if(vers.value())
    {
      print_version();
      cout << endl;
      exit(0);
    } 
  if(copyright.value())
    {
      print_copyright();
      exit(0);
    } 
  if(! options.check_compulsory_arguments())
    {
      print_version();
      cout << endl;
      cout << "Usage: melodic <options> -i <filename>" << endl <<
	"Please specify input file name correctly or try melodic --help" 
	   << endl << endl;
      exit(2);
    }     

  // check for invalid values
  if (inputfname.value().size()<1) {
    cerr << "ERROR:: Input volume filename not found\n\n";
    print_usage(argc,argv);
    exit(2);
  }
  if (approach.value() != "symm" && approach.value() != "defl"  && 
      approach.value() != "jade" && approach.value() != "maxent"){
    cerr << "ERROR:: unknown approach \n\n";
    print_usage(argc,argv);
    exit(2);
  }
  if (nonlinearity.value() != "pow3" && nonlinearity.value() != "tanh" 
      && nonlinearity.value() != "gauss" ){
    cerr << "ERROR:: unknown nonlinearity \n\n";
    print_usage(argc,argv);
    exit(2);
  }
  if (maxNumItt.value() < 1){
    cerr << "ERROR:: maxItt too small \n\n";
    print_usage(argc,argv);
    exit(2);
  }
  if (epsilon.value() < 0.000000001){
    cerr << "ERROR:: epsilon too small  \n\n";
    print_usage(argc,argv);
    exit(2);
  }
  if (epsilon.value() >= 0.1){
    cerr << "ERROR:: epsilon too large  \n\n";
    print_usage(argc,argv);
    exit(2);
  }
  if (nlconst1.value() <= 0){
    cerr << "ERROR:: nlconst1 negative  \n\n";
    print_usage(argc,argv);
    exit(2);
  }
  if (!remove_meanvol.value()){
    varnorm.set_T(false);
  }
  if (threshold.value()<=0){
    use_mask.set_T(false);
  }
  if(threshold.value()>=1){
    threshold.set_T(threshold.value()/100);
  }
  if (nlconst2.value() <= 0){
    cerr << "ERROR:: nlconst2 negative  \n\n";
    print_usage(argc,argv);
    exit(2);
  }
  if (dummy.value() < 0){
    cerr << "ERROR:: negative dummy value  \n\n";
    print_usage(argc,argv);
    exit(2);
  }
  if (repeats.value() < 1){
    cerr << "ERROR:: repeats < 1 \n\n";
    print_usage(argc,argv);
    exit(2);
  }
  
  //transform inputfname to its basename
  string basename = inputfname.value();
  make_basename(basename);
  inputfname.set_T(basename);

  //create melodic directory name
  if(logdir.value()==string("melodic.log")){
    logdir.set_T(string(inputfname.value()+".ica"));
  }

  // setup logger directory
  logger.makeDir(logdir.value());

  message(endl << "Melodic Version " << version << endl << endl);

  // parse again so that options are logged
  for(int a = 0; a < argc; a++)
    logger.str() << argv[a] << " ";
  logger.str() << endl << "---------------------------------------------" << endl << endl;

  message("Melodic results will be in " << logger.getDir() << endl << endl);
}

void MelodicOptions::print_usage(int argc, char *argv[])
{
  print_version();
  options.usage();
  /* cout << "Usage: " << argv[0] << " "; 

    Have own usage output here
  */
}

void MelodicOptions::print_version()
{
  cout << "MELODIC Version " << version;
}

void MelodicOptions::print_copyright()
{
  print_version();
  cout << endl << "Copyright (C) 1999-2001 University of Oxford " << endl;
}

void MelodicOptions::status()
{
  cout << " version = " << version << endl;
 
  cout << " logdir = "  << logdir.value() << endl;
  cout << " inputfname = "  << inputfname.value() << endl;
  cout << " outputfname = "  << outputfname.value() << endl;
  cout << " guessfname = "  << guessfname.value() << endl;
  cout << " maskfname = "  << maskfname.value() << endl;
  cout << " paradigmfname = "  <<  paradigmfname.value() << endl;
  cout << " nonlinearity = "  << nonlinearity.value() << endl;
  cout << " approach = "  << approach.value() << endl;
  
  cout << " pca_dim = "  << pca_dim.value() << endl;
  cout << " segment = "  << segment.value() << endl;
  cout << " numICs = "  << numICs.value() << endl;
  cout << " maxNumItt = "  << maxNumItt.value() << endl;
  cout << " maxRestart = "  << maxRestart.value() << endl;
  cout << " dummy = "  << dummy.value() << endl;
  cout << " repeats = "  << repeats.value() << endl;
  cout << " nlconst1 = "  << nlconst1.value() << endl;
  cout << " nlconst2 = "  << nlconst2.value() << endl;
  cout << " epsilon = "  << epsilon.value() << endl;
  cout << " threshold = "  << threshold.value() << endl;
  
  cout << " help = "  << help.value() << endl;
  cout << " verbose = "  << verbose.value() << endl;
  cout << " vers = "  << vers.value() << endl;
  cout << " copyright = "  << copyright.value() << endl;
  cout << " perf_bet = "  << perf_bet.value() << endl;
  cout << " remove_meanvol = "  << remove_meanvol.value() << endl;
  cout << " remove_endslices  = " << remove_endslices.value() << endl;
  cout << " output_pca = "  << output_pca.value() << endl;
  cout << " output_white = "  << output_white.value() << endl;
  cout << " output_mask = "  << output_mask.value() << endl;
  cout << " output_mean = "  << output_mean.value() << endl;
  cout << " use_mask = "  << use_mask.value() << endl;
  cout << " output_unmix = "  << output_unmix.value() << endl;
  cout << " guess_remderiv = "  << guess_remderiv.value() << endl;
  cout << " filter  = " << filter.value() << endl;
  cout << " logPower  = " << logPower.value() << endl;
}

}
