/* {{{ Copyright etc. */

/*  slicer - produce any slicings of a 3D image

    Stephen Smith, FMRIB Image Analysis Group

    Copyright (C) 1999-2001 University of Oxford  */

/*  Part of FSL - FMRIB's Software Library
    WWW:      http://www.fmrib.ox.ac.uk/fsl
    Email:    fsl@fmrib.ox.ac.uk
    
    Developed at FMRIB (Oxford Centre for Functional Magnetic Resonance
    Imaging of the Brain), Department of Clinical Neurology, Oxford
    University, Oxford, UK
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or (at
    your option) any later version.
    
    This program is distributed in the hope that it will be useful, but in
    order that the University as a charitable foundation protects its
    assets for the benefit of its educational and research purposes, the
    University makes clear that no condition is made or to be implied, nor
    is any warranty given or to be implied, as to the accuracy of FSL, or
    that it will be suitable for any particular purpose or for use under
    any specific conditions.  Furthermore, the University disclaims all
    responsibility for the use which is made of FSL.  See the GNU General
    Public License for more details.
    
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
    USA */

/* }}} */
/* {{{ defines, includes and typedefs */

#include "libss/libss.h"
#include "libss/libavw.h"
#include "libss/libpicio.h"

int x_size, y_size, z_size, size, x_size_pic, y_size_pic, z_size_pic, nlut=0,
  compare=0, trans=-10;
unsigned char *picr, *picg, *picb;	  
image_struct im, im2;
FDT *imr, *img, *imb;

/* }}} */
/* {{{ usage */

void usage(void)
{
  printf("\nUsage: slicer <input> [input2] [main options] [output options - any number of these]\n\n");

  printf("Main options: [-l <lut>] [-s <scale>] [-i <intensitymin> <intensitymax>] [-t]\n");
  printf("These must be before output options and must be in the above order.\n");
  printf("-l <lut> : use a different colour map from that specified in the header.\n");
  printf("-t       : produce semi-transparent (dithered) edges.\n\n");

  printf("Output options:\n");
  printf("[-x/y/z <slice> <filename>]      : output sagittal, coronal or axial slice\n     (if <slice> >0 it is a fraction of image dimension, if <0, it is an absolute slice number)\n");
  printf("[-a <filename>]                  : output mid-sagittal, -coronal and -axial slices into one image\n");
  printf("[-A <width> <filename>]          : output _all_ axial slices into one image of _max_ width <width>\n");
  printf("[-S <sample> <width> <filename>] : as -A but only include every <sample>'th slice\n\n");

  exit(1);
}

/* }}} */
/* {{{ write sagittal slice */

void sag(float xx, int p, int width)
{
  float yy, zz;
  int   y, z;

  if (xx<0)
    xx=-xx;
  else
    xx*=(float)x_size;
  xx=MIN(x_size-1.0001,xx);

  for(y=0; y<y_size_pic; y++)
    for(z=0; z<z_size_pic; z++)
      {
	yy=MIN(y_size-1.0001,y/im.yv);
	zz=MIN(z_size-1.0001,z/im.zv);

	if (nlut==0) {
	  picr[p+(z_size_pic-1-z)*width+y] = MIN(255,MAX(0,TLI(im,xx,yy,zz)));

	  if (compare)
	    {
	      picg[p+(z_size_pic-1-z)*width+y] = picr[p+(z_size_pic-1-z)*width+y];
	      picb[p+(z_size_pic-1-z)*width+y] = picr[p+(z_size_pic-1-z)*width+y];
	      if ( (IA(im2.i,(int)(xx+0.5),(int)(yy+0.5),(int)(zz+0.5))>0) &&
		   ((y+z)%2>trans) )
		{
		  picr[p+(z_size_pic-1-z)*width+y] = 255;
		  picg[p+(z_size_pic-1-z)*width+y] = picb[p+(z_size_pic-1-z)*width+y] = 0;
		}
	    }
	} else {
	  im.i=imr;
	  picr[p+(z_size_pic-1-z)*width+y]= MIN(255,MAX(0,TLI(im,xx,yy,zz)));
	  im.i=img;
	  picg[p+(z_size_pic-1-z)*width+y]= MIN(255,MAX(0,TLI(im,xx,yy,zz)));
	  im.i=imb;
	  picb[p+(z_size_pic-1-z)*width+y]= MIN(255,MAX(0,TLI(im,xx,yy,zz)));
	}
      }
}

/* }}} */
/* {{{ write coronal slice */

void cor(float yy, int p, int width)
{
float xx, zz;
int   x, z;

  if (yy<0)
    yy=-yy;
  else
    yy*=(float)y_size;
  yy=MIN(y_size-1.0001,yy);

  for(x=0; x<x_size_pic; x++)
    for(z=0; z<z_size_pic; z++)
      {
	xx=MIN(x_size-1.0001,x/im.xv);
	zz=MIN(z_size-1.0001,z/im.zv);

	if (nlut==0) {
	  picr[p+(z_size_pic-1-z)*width+x] = MIN(255,MAX(0,TLI(im,xx,yy,zz)));
	  
	  if (compare)
	    {
	      picg[p+(z_size_pic-1-z)*width+x] = picr[p+(z_size_pic-1-z)*width+x];
	      picb[p+(z_size_pic-1-z)*width+x] = picr[p+(z_size_pic-1-z)*width+x];
	      if ( (IA(im2.i,(int)(xx+0.5),(int)(yy+0.5),(int)(zz+0.5))>0) &&
		   ((x+z)%2>trans) )
		{
		  picr[p+(z_size_pic-1-z)*width+x] = 255;
		  picg[p+(z_size_pic-1-z)*width+x] = picb[p+(z_size_pic-1-z)*width+x] = 0;
		}
	    }
	} else {
	  im.i=imr;
	  picr[p+(z_size_pic-1-z)*width+x]= MIN(255,MAX(0,TLI(im,xx,yy,zz)));
	  im.i=img;
	  picg[p+(z_size_pic-1-z)*width+x]= MIN(255,MAX(0,TLI(im,xx,yy,zz)));
	  im.i=imb;
	  picb[p+(z_size_pic-1-z)*width+x]= MIN(255,MAX(0,TLI(im,xx,yy,zz)));
	}
      }
}

/* }}} */
/* {{{ write axial slice */

void axi(float zz, int p, int width)
{
  float xx, yy;
  int   x, y;

  if (zz<0)
    zz=-zz;
  else
    zz*=(float)z_size;
  zz=MAX(0,MIN(z_size-1.0001,zz));

  for(x=0; x<x_size_pic; x++)
    for(y=0; y<y_size_pic; y++)
      {
	xx=MIN(x_size-1.0001,x/im.xv);
	yy=MIN(y_size-1.0001,y/im.yv);

	if (nlut==0) {
	  picr[p+(y_size_pic-1-y)*width+x] = MIN(255,MAX(0,TLI(im,xx,yy,zz)));
	  
	  if (compare)
	    {
	      picg[p+(y_size_pic-1-y)*width+x] = picr[p+(y_size_pic-1-y)*width+x];
	      picb[p+(y_size_pic-1-y)*width+x] = picr[p+(y_size_pic-1-y)*width+x];
	      if ( (IA(im2.i,(int)(xx+0.5),(int)(yy+0.5),(int)(zz+0.5))>0) &&
		   ((x+y)%2>trans) )
		{
		  picr[p+(y_size_pic-1-y)*width+x] = 255;
		  picg[p+(y_size_pic-1-y)*width+x] = picb[p+(y_size_pic-1-y)*width+x] = 0;
		}
	    }
	} else {
	  im.i=imr;
	  picr[p+(y_size_pic-1-y)*width+x]= MIN(255,MAX(0,TLI(im,xx,yy,zz)));
	  im.i=img;
	  picg[p+(y_size_pic-1-y)*width+x]= MIN(255,MAX(0,TLI(im,xx,yy,zz)));
	  im.i=imb;
	  picb[p+(y_size_pic-1-y)*width+x]= MIN(255,MAX(0,TLI(im,xx,yy,zz)));
	}
      }
}

/* }}} */
/* {{{ write_pnm */

void write_pnm(char *fname, int width, int height)
{
  if ( (nlut>0) || (compare) )
    write_ppm(fname,width,height,picr,picg,picb);
  else
    write_pgm(fname,width,height,picr);
}

/* }}} */
/* {{{ main */

int main(argc, argv)
  int   argc;
  char  *argv [];
{
  /* {{{ vars */

char   lut[10000];
int    i, argindex=1, picsize, rlut[100000], glut[100000], blut[100000];
float  tmpf;
FDT    *edge;

/* }}} */

  if (argc<2)
    usage();

  /* {{{ read image */

avw_read(argv[argindex++],&im);

x_size = im.x;
y_size = im.y;
z_size = im.z;

size = x_size * y_size * z_size;

/* }}} */
  /* {{{ read second (optional) image */

if ((char)argv[argindex][0]!='-')
{
  int x, y, z;
  unsigned char *orient;
  float *tmpedge;
  FDT ethresh;

  compare=1;

  avw_read(argv[argindex++],&im2);
 
  if ( (x_size!=im2.x) || (y_size!=im2.y) || (z_size!=im2.z) )
    {
      printf("Images aren't the same size - duh!\n");
      return 1;
    }

  if ( (edge=(FDT*)calloc(size,sizeof(FDT)))==NULL )
    {
      printf("edge calloc failed!\n");
      exit(1);
    }
  if ( (tmpedge=(float*)calloc(size,sizeof(float)))==NULL)
    {
      printf("tmpedge calloc failed!\n");
      exit(1);
    }
  orient=(unsigned char*)malloc(size);

  for(z=1; z<z_size-1; z++)
    for(y=1; y<y_size-1; y++)
      for(x=1; x<x_size-1; x++)
	{
	  double xx=IA(im2.i,x+1,y,z)-IA(im2.i,x-1,y,z),
	    yy=IA(im2.i,x,y+1,z)-IA(im2.i,x,y-1,z),
	    zz=IA(im2.i,x,y,z+1)-IA(im2.i,x,y,z-1);

	  xx*=xx; yy*=yy; zz*=zz;

	  if ( (xx>yy) && (xx>zz) )
	    IA(orient,x,y,z)=0;
	  else
	    {
	      if (yy>zz)
		IA(orient,x,y,z)=1;
	      else
		IA(orient,x,y,z)=2;
	    }

	  IA(tmpedge,x,y,z) = xx+yy+zz;
	}

  for(z=1; z<z_size-1; z++)
    for(y=1; y<y_size-1; y++)
      for(x=1; x<x_size-1; x++)
	{
	  if (IA(orient,x,y,z)==0)
	    {
	      if ( (IA(tmpedge,x,y,z)>IA(tmpedge,x-1,y,z)) && (IA(tmpedge,x,y,z)>=IA(tmpedge,x+1,y,z)) )
		IA(edge,x,y,z)=sqrt(IA(tmpedge,x,y,z))/12;
	    }
	  else
	    {
	      if (IA(orient,x,y,z)==1)
		{
		  if ( (IA(tmpedge,x,y,z)>IA(tmpedge,x,y-1,z)) && (IA(tmpedge,x,y,z)>=IA(tmpedge,x,y+1,z)) )
		    IA(edge,x,y,z)=sqrt(IA(tmpedge,x,y,z))/12;
		}
	      else
		{
		  if ( (IA(tmpedge,x,y,z)>IA(tmpedge,x,y,z-1)) && (IA(tmpedge,x,y,z)>=IA(tmpedge,x,y,z+1)) )
		    IA(edge,x,y,z)=sqrt(IA(tmpedge,x,y,z))/12;
		}
	    }
	}

  free(tmpedge);
  free(orient);

  im2.i=edge;
  im2.min=im2.max=0;
  find_thresholds(&im2,0.1);
  ethresh=((double)im2.thresh98-(double)im2.thresh2)*0.5+im2.thresh2;
  /*printf("%f %f %f\n",(float)im2.thresh2,(float)im2.thresh98,(float)ethresh);*/

  for(z=1; z<z_size-1; z++)
    for(y=1; y<y_size-1; y++)
      for(x=1; x<x_size-1; x++)
	if (IA(edge,x,y,z)<ethresh)
	  IA(edge,x,y,z)=0;
}

/* }}} */
  /* {{{ test for lut option and read lut */

sprintf(lut,"%s/etc/luts/%s.lut",getenv("FSLDIR"),im.lut);

if (!strcmp(argv[argindex],"-l"))
{
  strcpy(lut,argv[argindex+1]);
  argindex+=2;
}

{
  FILE *tmpfp;
  char lutline[10000];
  float fr, fg, fb;

  if ((tmpfp=fopen(lut,"r"))!=NULL)
    {
      while (fgets(lutline, 10000, tmpfp)!=NULL)
	if ( (strncmp(lutline,"<-color{",8)==0) )
	  {
	    sscanf(lutline+8,"%f,%f,%f",&fr,&fg,&fb);
            rlut[nlut]=MAX(0,MIN(255,255*fr));
            glut[nlut]=MAX(0,MIN(255,255*fg));
            blut[nlut]=MAX(0,MIN(255,255*fb));
	    /*printf("%d %d %d\n",rlut[nlut],glut[nlut],blut[nlut]);*/
	    nlut++;
	  }
    }
}

/* }}} */
  /* {{{ set scale etc */

tmpf = MIN(im.xv,MIN(im.yv,im.zv));

if (!strcmp(argv[argindex],"-s"))
{
  tmpf /= atof(argv[argindex+1]);
  argindex+=2;
}

im.xv/=tmpf;
im.yv/=tmpf;
im.zv/=tmpf;

x_size_pic = x_size*im.xv;
y_size_pic = y_size*im.yv;
z_size_pic = z_size*im.zv;

/*printf("%f %f %f %d %d %d\n",im.xv,im.yv,im.zv,x_size_pic,y_size_pic,z_size_pic);*/

/* }}} */
  /* {{{ find thresholds */

if (nlut==0) /* otherwise (if using luts) then the display range should already have been set correctly */
{
  im.min=im.max=0;
  find_thresholds(&im,0.1);

  /*printf("Thresholds: %f %f %f %f %f\n",(float)im.min,(float)im.thresh2,(float)im.thresh,(float)im.thresh98,(float)im.max);*/

  /* fix thresholds if a simple binary image */
  if (im.thresh2==im.thresh98)
    {
      im.thresh2=im.min;
      im.thresh98=im.max;
    }
} else {
    im.thresh2  = im.min;
    im.thresh98 = im.max;
}

if (!strcmp(argv[argindex],"-i"))
{
  im.thresh2  = atof(argv[argindex+1]);
  im.thresh98 = atof(argv[argindex+2]);
  argindex+=3;
}

/* }}} */
  /* {{{ rescale image intensity */

/* now rescale image */
if (nlut==0)
{
  for(i=0;i<size;i++)
    im.i[i] = ( ( ((double)im.i[i]) - ((double)im.thresh2) ) * 255.001 /
		( ((double)im.thresh98) - ((double)im.thresh2) ) );
} else {
  imr=malloc(size*sizeof(FDT));
  img=malloc(size*sizeof(FDT));
  imb=malloc(size*sizeof(FDT));

  for(i=0;i<size;i++)
    {
      int tmp = ( ( ((double)im.i[i]) - ((double)im.thresh2) ) * (nlut-1) /
		  ( ((double)im.thresh98) - ((double)im.thresh2) ) );
      tmp = MIN(nlut-1,MAX(0,tmp));
      
      imr[i] = rlut[tmp];
      img[i] = glut[tmp];
      imb[i] = blut[tmp];
    }
}

/* }}} */
  /* {{{ transparent edges? */

if (!strcmp(argv[argindex],"-t"))
{
  trans=0;
  argindex++;
}

/* }}} */
  /* {{{ malloc pics */

picsize = MAX(x_size_pic,MAX(y_size_pic,z_size_pic));
picsize = picsize*picsize*MAX(10,2*z_size);

/*printf("%d %d %d %d\n",x_size_pic,y_size_pic,z_size_pic,picsize);*/

picr = malloc(picsize);
picg = malloc(picsize);
picb = malloc(picsize);

/* }}} */
  /* {{{ process slice requests */

for (i = argindex; i < argc; i++) {

  /* {{{ setup some dimensions and zero images */

int sagcorskip=0, axiskip=0,
  height=MAX(y_size_pic,z_size_pic);

if (y_size_pic>z_size_pic)
  sagcorskip=(y_size_pic-z_size_pic)/2;
else
  axiskip=(z_size_pic-y_size_pic)/2;

memset(picr,0,picsize);
memset(picg,0,picsize);
memset(picb,0,picsize);

/* }}} */

  if (!strcmp(argv[i], "-a"))
    /* {{{ write default slices */

{
  int width=y_size_pic+2*x_size_pic;

  sag(0.5,                         sagcorskip*width, width);
  cor(0.5, y_size_pic            + sagcorskip*width, width);
  axi(0.5, y_size_pic+x_size_pic + axiskip*width,    width);

  write_pnm(argv[++i],width,height);
}

/* }}} */
  else if ( (!strcmp(argv[i], "-A")) || (!strcmp(argv[i], "-S")) )
    /* {{{ write all slices */

{
  int maxwidth, nx, ny, width, row, column, height, z=0, sample=1, slices;

  if (!strcmp(argv[i], "-S"))
    sample=atoi(argv[++i]);

  maxwidth=atoi(argv[++i]);

  slices=ceil( ((float)z_size) / sample);
  nx=MIN(MAX(maxwidth/x_size_pic,1),slices);
  ny=ceil( ((float)slices) / nx);
  width=x_size_pic*nx;
  height=y_size_pic*ny;
  /*printf("%d %d %d %d %d\n",slices,nx,ny,width,height);*/
    
  for(row=0;row<ny;row++)
    for(column=0;column<nx;column++)
      {
	if (z<z_size)
	  axi(-z, row*y_size_pic*width + x_size_pic*column, width);
	z+=sample;
      }

  write_pnm(argv[++i],width,height);
}

/* }}} */
  else if (!strcmp(argv[i], "-x"))
    /* {{{ write sagittal slice */

{
  sag(atof(argv[++i]), sagcorskip*y_size_pic, y_size_pic);
  write_pnm(argv[++i],y_size_pic,height);
}

/* }}} */
  else if (!strcmp(argv[i], "-y"))
    /* {{{ write coronal slice */

{
  cor(atof(argv[++i]), sagcorskip*x_size_pic, x_size_pic);
  write_pnm(argv[++i],x_size_pic,height);
}

/* }}} */
  else if (!strcmp(argv[i], "-z"))
    /* {{{ write axial slice */

{
  axi(atof(argv[++i]), axiskip*x_size_pic, x_size_pic);
  write_pnm(argv[++i],x_size_pic,height);
}

/* }}} */

  else
    usage();
}

/* }}} */

  return(0);
}

/* }}} */
